@@ -25,11 +25,14 @@
 ;; Utility functions used in the mu4e
 
 ;;; Code:
+(eval-when-compile (byte-compile-disable-warning 'cl-functions))
 (require 'cl)
+
 (require 'html2text)
 (require 'mu4e-vars)
+(require 'mu4e-about)
 (require 'doc-view)
-
+(require 'org) ;; for org-parse-time-string
 
 (defcustom mu4e-html2text-command nil
   "Shell command that converts HTML from stdin into plain text on
@@ -40,7 +43,6 @@ recommended you use \"html2text -utf8 -width 72\"."
   :group 'mu4e-view
   :safe 'stringp)
 
-
 (defcustom mu4e-view-prefer-html nil
   "Whether to base the body display on the HTML-version of the
 e-mail message (if there is any."
@@ -59,34 +61,15 @@ dir already existed, or has been created, nil otherwise."
       (mu4e~proc-mkdir dir))
     (t nil)))
 
-
 (defun mu4e-format (frm &rest args)
   "Create [mu4e]-prefixed string based on format FRM and ARGS."
   (concat "[" mu4e-logo "] "  (apply 'format frm args)))
 
 (defun mu4e-message (frm &rest args)
-  "Like `message', but prefixed with mu4e."
-  (message "%s" (apply 'mu4e-format frm args)))
-
-
-(defun mu4e~read-option-normalize-list (options)
-  "Turn a list OPTIONS into normal-form for `mu4e-read-option'."
-  ;; transform options into 'normal-form', so that in case an option has 'nil
-  ;; for CHAR, it's replaced by the first letter of OPTIONSTRING (and that char
-  ;; is eaten off OPTIONSTR. If RESULT is nil, replace it by CHAR
-  (map 'list
-    (lambda (option)
-      (if (nth 1 option)
-	(list
-	  (nth 0 option)
-	  (nth 1 option)
-	  (or (nth 2 option) (nth 1 option))) ;
-	(list
-	  (substring (nth 0 option) 1)    ;; chop off first char
-	  (string-to-char (nth 0 option)) ;; first char as shortcut
-	  (or (nth 2 option) (nth 1 option)
-	    (string-to-char (nth 0 option))))))
-    options))
+  "Like `message', but prefixed with mu4e. If we're waiting for
+user-input, don't show anyhting."
+  (unless (waiting-for-user-input-p)
+    (message "%s" (apply 'mu4e-format frm args))))
 
 (defun mu4e~read-char-choice (prompt choices)
   "Compatiblity wrapper for `read-char-choice', which is emacs-24
@@ -104,47 +87,49 @@ describes a multiple-choice question to the user, OPTIONS describe
 the options, and is a list of cells describing particular
 options. Cells have the following structure:
 
-   (OPTIONSTRING CHAR [RESULT])
+   (OPTIONSTRING . RESULT)
 
- where CHAR is a short-cut character for the
-option, and OPTIONSTRING is a non-empty string describing the
-option. If CHAR is nil or not-specified, the first character of the
-optionstring is used.
-
-If RESULT is provide, this will be returned if the user presses the
-corresponding CHAR; otherwise, CHAR is returned.
+ where OPTIONSTRING is a non-empty string describing the
+ option. The first character of OPTIONSTRING is used as the
+ shortcut, and obviously all shortcuts must be different, so you
+ can prefix the string with an uniquifying character.
 
 The options are provided as a list for the user to choose from;
 user can then choose by typing CHAR.  Example:
   (mu4e-read-option \"Choose an animal: \"
-              '((\"Monkey\" ?m) (\"Gnu\" ?g) (\"platipus\")))
-User now will be presented with a list:
-   \"Choose an animal: [m]Monkey, [g]Gnu, [p]latipus\"."
-  (let* ((options (mu4e~read-option-normalize-list options)) 
-	  (prompt (mu4e-format "%s" prompt))
+              '((\"Monkey\" . monkey) (\"Gnu\" . gnu) (\"xMoose\" . moose)))
+
+User now will be presented with a list: \"Choose an animal:
+   [M]onkey, [G]nu, [x]Moose\".
+
+Function will return the cdr of the list element."
+  (let* ((prompt (mu4e-format "%s" prompt))
 	  (chosen)
 	  (optionsstr
 	    (mapconcat
 	      (lambda (option)
-		(let* ((descr (car option))
-		      (kar (and (cdr option) (cadr option))))
-		;; handle the empty kar case
-		(unless kar
-		  (setq ;; eat first kar from descr; use it as kar
-		    kar   (string-to-char descr)
-		    descr (substring descr 1)))
-		(concat
-		  "[" (propertize (make-string 1 kar)
-			'face 'mu4e-highlight-face) "]"
-		  descr))) options ", "))
+		;; try to detect old-style options, and warn
+		(when (characterp (car-safe (cdr-safe option)))
+		  (error (concat "Please use the new format for options/actions; "
+			   "see the manual")))
+		(let* ((kar (substring (car option) 0 1))
+			(val (cdr option)))
+		  (concat
+		    "[" (propertize kar 'face 'mu4e-highlight-face) "]"
+		    (substring (car option) 1))))
+	      options ", "))
 	  (response
 	    (mu4e~read-char-choice
 	      (concat prompt optionsstr
 		" [" (propertize "C-g" 'face 'mu4e-highlight-face) " to quit]")
-	      (map 'list (lambda(elm) (nth 1 elm)) options))) ;; the allowable chars
+	      ;; the allowable chars
+	      (map 'list (lambda(elm) (string-to-char (car elm))) options)))
 	  (chosen
-	    (find-if (lambda (option) (eq response (nth 1 option))) options)))
-    (nth 2 chosen))) 
+	    (find-if
+	      (lambda (option) (eq response (string-to-char (car option))))
+	      options)))
+    (unless chosen (error "%S not found" response))
+    (cdr chosen)))
 
 
 (defun mu4e~get-maildirs-1 (path &optional mdir)
@@ -170,18 +155,23 @@ paths."
 	  (add-to-list 'maildirs (if mdir mdir "/") t)))
       (setq maildirs (append maildirs
 		       (mu4e~get-maildirs-1 path (concat mdir "/" dir)))))
-    maildirs)) 
+    maildirs))
 
+(defvar mu4e~maildir-list nil "Cached list of maildirs.")
 
 (defun mu4e-get-maildirs (path)
   "Get maildirs under path, recursively, as a list of relative
 paths (ie., /archive, /sent etc.). Most of the work is done in
-`mu4e-get-maildirs-1'."
-  (sort (mu4e~get-maildirs-1 path)
-    (lambda (m1 m2)
-      (when (string= m1 "/")
-	-1 ;; '/' comes first
-	(compare-strings m1 0 nil m2 0 nil t)))))
+`mu4e-get-maildirs-1'. Note, these results are /cached/, so the
+list of maildirs will not change until you restart mu4e."
+  (unless mu4e~maildir-list
+    (setq mu4e~maildir-list
+      (sort (mu4e~get-maildirs-1 path)
+	(lambda (m1 m2)
+	  (when (string= m1 "/")
+	    -1 ;; '/' comes first
+	    (compare-strings m1 0 nil m2 0 nil t))))))
+  mu4e~maildir-list)
 
 
 (defun mu4e-ask-maildir (prompt)
@@ -211,7 +201,7 @@ maildirs under `mu4e-maildir."
 	  (ido-completing-read prompt (mu4e-get-maildirs mu4e-maildir))
 	  (or (car-safe
 		(find-if (lambda (item) (= kar (cdr item))) mu4e-maildir-shortcuts))
-	    (error "Invalid shortcut '%c'" kar))))))) 
+	    (error "Invalid shortcut '%c'" kar)))))))
 
 
 (defun mu4e-ask-maildir-check-exists (prompt)
@@ -224,7 +214,7 @@ and offer to create it if it does not exist yet."
 	     (mu4e-format "%s does not exist. Create now?" fullpath))
 	      (mu4e~proc-mkdir fullpath)))
     mdir))
-   
+
 
 (defun mu4e-mark-for-move-set (&optional target)
   "Mark message at point or, if region is active, all messages in
@@ -273,7 +263,7 @@ KAR, or raise an error if none is found."
    (if chosen-bm
      (nth 0 chosen-bm)
      (error "Invalid shortcut '%c'" kar))))
- 
+
 
 ;;; converting flags->string and vice-versa ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
 (defun mu4e~flags-to-string-raw (flags)
@@ -340,8 +330,6 @@ http://cr.yp.to/proto/maildir.html "
 ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
 
 
-
-
 (defun mu4e-display-size (size)
   "Get a string representation of SIZE (in bytes)."
   (cond
@@ -384,9 +372,20 @@ function prefers the text part, but this can be changed by setting
 		  (buffer-string)))
 	      (t ;; otherwise, an empty body
 		""))))
-    ;; and finally, remove some crap from the remaining string.
-    (replace-regexp-in-string "[
- ]" " " body nil nil nil)))
+    ;; and finally, remove some crap from the remaining string; it seems
+    ;; esp. outlook lies about its encoding (ie., it says 'iso-8859-1' but
+    ;; really it's 'windows-1252'), thus giving us these funky chars. here, we
+    ;; either remove them, or replace with 'what-was-meant' (heuristically)
+    (with-temp-buffer
+      (insert body)
+      (goto-char (point-min))
+      (while (re-search-forward "[
+ ]" nil t)
+	(replace-match
+	  (cond
+	    ((string= (match-string 0) "") "'")
+	    (t		                       ""))))
+      (buffer-string))))
 
 
 
@@ -400,7 +399,7 @@ top level if there is none."
 	  ('mu4e-view-mode "(mu4e)Message view")
 	  (t               "mu4e"))))
 
- 
+
 
 ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
 (defun mu4e-msg-field (msg field)
@@ -429,7 +428,6 @@ A message plist looks something like:
  :in-reply-to \"6BDC23465F79238203498230942D81EE81AF0114E4E74@123213.mail.example.com\"
  :body-txt \"Hi Tom, ...\"
 \)).
-
 Some  notes on the format:
 - The address fields are lists of pairs (NAME . EMAIL), where NAME can be nil.
 - The date is in format emacs uses in `current-time'
@@ -466,7 +464,7 @@ point in eiter the headers buffer or the view buffer."
   "Get the most recent query or nil if there is none."
   (when (buffer-live-p mu4e~headers-buffer)
     (with-current-buffer mu4e~headers-buffer
-      mu4e~headers-query)))
+      mu4e~headers-last-query)))
 
 (defun mu4e-select-other-view ()
   "When the headers view is selected, select the message view (if
@@ -484,7 +482,7 @@ that has a live window), and vice versa."
       (mu4e-message "No window to switch to"))))
 
 
-(defconst mu4e-output-buffer-name "*mu4e-output"
+(defconst mu4e-output-buffer-name "*mu4e-output*"
   "*internal* Name of the mu4e output buffer.")
 
 (defun mu4e-process-file-through-pipe (path pipecmd)
@@ -506,7 +504,7 @@ that has a live window), and vice versa."
 process."
   (let ((type (plist-get info :info)))
     (cond
-      ((eq type 'add) t) ;; do nothing	
+      ((eq type 'add) t) ;; do nothing
       ((eq type 'index)
 	(if (eq (plist-get info :status) 'running)
 	  (mu4e-message "Indexing... processed %d, updated %d"
@@ -527,9 +525,11 @@ process."
 ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
 
 
-(defconst mu4e-update-buffer-name "*mu4e-update*"
-  "*internal* Name of the buffer for message retrieval / database
-  updating.")
+(defconst mu4e~update-buffer-name "*mu4e-update*"
+  "Name of the buffer for message retrieval/database updating.")
+
+(defconst mu4e~update-buffer-height 8
+  "Height of the mu4e message retrieval/update buffer.")
 
 (defun mu4e-update-mail-show-window ()
   "Try to retrieve mail (using the user-provided shell command),
@@ -538,9 +538,13 @@ split-window."
   (interactive)
   (unless mu4e-get-mail-command
     (error "`mu4e-get-mail-command' is not defined"))
-  (let ((buf (get-buffer-create mu4e-update-buffer-name))
-	 (win
-	   (split-window (selected-window)
+  ;; delete any old update buffer
+  (when (buffer-live-p mu4e~update-buffer-name)
+    (with-current-buffer mu4e~update-buffer-name
+      (kill-buffer-and-window)))
+  ;; create a new one
+  (let ((buf (get-buffer-create mu4e~update-buffer-name))
+	 (win (split-window (selected-window)
 	     (- (window-height (selected-window)) 8))))
     (with-selected-window win
       (switch-to-buffer buf)
@@ -550,11 +554,30 @@ split-window."
       (mu4e-update-mail buf))))
 
 
-
-
 
 ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
 ;; start and stopping
+(defun mu4e~fill-contacts (contacts)
+  "We receive a list of contacts, which each contact of the form
+     (:name NAME :mail EMAIL)
+and fill the list `mu4e~contacts-for-completion' with it, with
+each element looking like
+    name <email>
+This is used by the completion function in mu4e-compose."
+  (let ((lst))
+    (dolist (contact contacts)
+      (let ((name (plist-get contact :name))
+	     (mail (plist-get contact :mail)))
+	(when mail
+	  (unless ;; ignore some address ('noreply' etc.)
+	    (and mu4e-compose-complete-ignore-address-regexp
+	      (string-match mu4e-compose-complete-ignore-address-regexp mail))
+	  (add-to-list 'lst
+	    (if name (format "%s <%s>" name mail) mail))))))
+    (setq mu4e~contacts-for-completion lst)
+    (mu4e-message "Contacts received: %d"
+      (length mu4e~contacts-for-completion))))
+
 
 (defun mu4e~check-requirements ()
   "Check for the settings required for running mu4e."
@@ -580,7 +603,7 @@ split-window."
 	(error "%S must start with a '/'" dir))
       (unless (mu4e-create-maildir-maybe path)
 	(error "%s (%S) does not exist" path var)))))
- 
+
 
 (defun mu4e~start (&optional func)
   "If mu4e is already running, execute function FUNC (if non-nil). Otherwise,
@@ -596,6 +619,7 @@ FUNC (if non-nil) afterwards."
       ;; better to check for specific features
       (unless (>= emacs-major-version 23)
 	(error "Emacs >= 23.x is required for mu4e"))
+
       ;; set up the 'pong' handler func
       (lexical-let ((func func))
 	(setq mu4e-pong-func
@@ -610,19 +634,34 @@ FUNC (if non-nil) afterwards."
 		  0 mu4e-update-interval 'mu4e-update-mail)))
 	    (mu4e-message "Started mu4e with %d message%s in store"
 	      doccount (if (= doccount 1) "" "s")))))
-	  ;; send the ping
-	  (mu4e~proc-ping)))
+      ;; send the ping
+      (mu4e~proc-ping)
+      ;; get the address list
+      (when mu4e-compose-complete-addresses
+	(setq mu4e-contacts-func 'mu4e~fill-contacts)
+	(mu4e~proc-contacts
+	  mu4e-compose-complete-only-personal
+	  (when mu4e-compose-complete-only-after
+	    (float-time
+	      (apply 'encode-time
+		(org-parse-time-string mu4e-compose-complete-only-after))))))))
 
 (defun mu4e~stop ()
-  "Quit the mu4e session."
-  (interactive)
-  (when (y-or-n-p (mu4e-format "Are you sure you want to quit?"))
-    (message nil)
-    (when mu4e-update-timer
-      (cancel-timer mu4e-update-timer)
-      (setq mu4e-update-timer nil))
-    (mu4e~proc-kill)
-    (kill-buffer)))
+  "Stop the mu4e session."
+  (when mu4e-update-timer
+    (cancel-timer mu4e-update-timer)
+    (setq
+      mu4e-update-timer nil
+      mu4e~maildir-list nil))
+  (mu4e~proc-kill)
+  ;; kill all main/view/headers buffer
+  (mapcar
+    (lambda (buf)
+      (with-current-buffer buf
+	(when (member major-mode '(mu4e-headers-mode mu4e-view-mode mu4e-main-mode))
+	  (kill-buffer))))
+    (buffer-list)))
+
 
 (defvar mu4e-update-timer nil
   "*internal* The mu4e update timer.")
@@ -649,15 +688,15 @@ processing takes part in the background, unless buf is non-nil."
 		;; sadly, fetchmail returns '1' when there is no mail; this is
 		;; not really an error of course, but it's hard to distinguish
 		;; from a genuine error
-		(maybe-error (or (not (eq status 'exit)) (/= code 0))))
+		(maybe-error (or (not (eq status 'exit)) (/= code 0)))
+		(buf (process-buffer proc)))
 	  (message nil)
 	  ;; there may be an error, give the user up to 5 seconds to check
 	  (when maybe-error
 	    (sit-for 5))
-	  (mu4e~proc-index mu4e-maildir)
-	  (let ((buf (process-buffer proc)))
-	    (when (buffer-live-p buf)
-	      (kill-buffer buf))))))))
+	  (mu4e~proc-index mu4e-maildir mu4e-my-email-addresses)
+	  (when (buffer-live-p buf)
+	    (kill-buffer buf)))))))
 
 
 
@@ -704,6 +743,8 @@ either 'to-server, 'from-server or 'misc. This function is meant for debugging."
 	      (beginning-of-line)
 	      (delete-region (point-min) (point)))))))))
 
+
+
 (defun mu4e-toggle-logging ()
   "Toggle between enabling/disabling debug-mode (in debug-mode,
 mu4e logs some of its internal workings to a log-buffer. See
@@ -715,6 +756,8 @@ mu4e logs some of its internal workings to a log-buffer. See
     (if mu4e-debug "enabled" "disabled"))
   (mu4e-log 'misc "logging enabled"))
 
+
+
 (defun mu4e-show-log ()
   "Visit the mu4e debug log."
   (interactive)
@@ -722,8 +765,113 @@ mu4e logs some of its internal workings to a log-buffer. See
     (unless (buffer-live-p buf)
       (error "No debug log available"))
     (switch-to-buffer buf)))
+
+
+(defun mu4e-split-ranges-to-numbers (str n)
+  "Convert STR containing attachment numbers into a list of numbers.
+STR is a string; N is the highest possible number in the list.
+This includes expanding e.g. 3-5 into 3,4,5.  If the letter
+\"a\" ('all')) is given, that is expanded to a list with numbers [1..n]."
+  (let ((str-split (split-string str))
+	 beg end list)
+    (dolist (elem str-split list)
+      ;; special number "a" converts into all attachments 1-N.
+      (when (equal elem "a")
+	(setq elem (concat "1-" (int-to-string n))))
+      (if (string-match "\\([0-9]+\\)-\\([0-9]+\\)" elem)
+	;; we have found a range A-B, which needs converting
+	;; into the numbers A, A+1, A+2, ... B.
+	(progn
+	  (setq beg (string-to-number (match-string 1 elem))
+	    end (string-to-number (match-string 2 elem)))
+	  (while (<= beg end)
+	    (add-to-list 'list beg 'append)
+	    (setq beg (1+ beg))))
+	;; else just a number
+	(add-to-list 'list (string-to-number elem) 'append)))
+    ;; Check that all numbers are valid.
+    (mapc
+      #'(lambda (x)
+	  (cond
+	    ((> x n)
+	      (error "Attachment %d bigger than maximum (%d)" x n))
+	    ((< x 1)
+	      (error "Attachment number must be greater than 0 (%d)" x))))
+      list)))
+
+
 ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
 
+(defvar mu4e-imagemagick-identify "identify"
+  "Name/path of the Imagemagick 'identify' program.")
+
+(defun mu4e-display-image (imgpath &optional maxwidth)
+  "Display image IMG at point; optionally specify
+MAXWIDTH. Function tries to use imagemagick if available (ie.,
+emacs was compiled with inmagemagick support); otherwise MAXWIDTH
+is ignored."
+  (let* ((have-im (and (fboundp 'imagemagick-types)
+		    (imagemagick-types))) ;; hmm, should check for specific type
+	  (identify (and have-im maxwidth
+		      (executable-find mu4e-imagemagick-identify)))
+	  (props (and identify (shell-command-to-string
+				 (format "%s -format '%%w' %s"
+				   identify (shell-quote-argument imgpath)))))
+	  (width (and props (string-to-number props)))
+	  (img (if have-im
+		 (if (> (or width 0) (or maxwidth 0))
+		   (create-image imgpath 'imagemagick nil :width maxwidth)
+		   (create-image imgpath 'imagemagick))
+		 (create-image imgpath))))
+    ;;(message "DISPLAY: %S %S" imgpath img)
+    (when img
+      (newline)
+      (insert-image img imgpath nil t))))
+
+
+
+(defun mu4e-hide-other-mu4e-buffers ()
+  "Bury mu4e-buffers (main, headers, view) (and delete all windows
+displaying it). Do _not_ bury the current buffer, though."
+  (interactive)
+  (let ((curbuf (current-buffer)))
+    ;; note: 'walk-windows' does not seem to work correctly when modifying
+    ;; windows; therefore, the doloops here
+    (dolist (frame (frame-list))
+      (dolist (win (window-list frame nil))
+	(with-current-buffer (window-buffer win)
+	  (unless (eq curbuf (current-buffer))
+	    (when (member major-mode '(mu4e-headers-mode mu4e-view-mode))
+	      (unless (one-window-p t)
+		(delete-window win))))))) nil t))
+
+
+(defun mu4e-get-time-date (prompt)
+  "Determine the emacs time value for the time/date entered by user
+  after PROMPT. Formats are all that are accepted by
+  `parse-time-string'."
+  (let ((timestr (read-string (mu4e-format "%s" prompt))))
+    (apply 'encode-time (org-parse-time-string timestr))))
+;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
+
+(defconst mu4e~main-about-buffer-name "*mu4e-about*"
+  "Name for the mu4e-about buffer.")
+
+(defun mu4e-about ()
+  "Show a buffer with the mu4e-about text."
+  (interactive)
+  (with-current-buffer
+    (get-buffer-create mu4e~main-about-buffer-name)
+    (let ((inhibit-read-only t))
+      (erase-buffer)
+      (insert mu4e-about)
+      (org-mode)))
+  (switch-to-buffer mu4e~main-about-buffer-name)
+  (setq buffer-read-only t)
+  (local-set-key "q" 'bury-buffer)
+  (goto-char (point-min)))
+;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
+
 
 (provide 'mu4e-utils)
 ;;; End of mu4e-utils.el
