@@ -26,30 +26,16 @@ struct _MugQueryBarPrivate {
 /* globals */
 static GtkContainerClass *parent_class = NULL;
 
+
 static guint signals[LAST_SIGNAL] = { 0 };
 
-GType
-mug_query_bar_get_type (void)
-{
-	static GType my_type = 0;
-	if (!my_type) {
-		static const GTypeInfo my_info = {
-			sizeof (MugQueryBarClass),
-			NULL,	/* base init */
-			NULL,	/* base finalize */
-			(GClassInitFunc) mug_query_bar_class_init,
-			NULL,	/* class finalize */
-			NULL,	/* class data */
-			sizeof (MugQueryBar),
-			0,	/* n_preallocs, ignored since 2.10 */
-			(GInstanceInitFunc) mug_query_bar_init,
-			NULL
-		};
-		my_type = g_type_register_static (GTK_TYPE_HBOX,
-						  "MugQueryBar", &my_info, 0);
-	}
-	return my_type;
-}
+
+#ifdef HAVE_GTK3
+G_DEFINE_TYPE (MugQueryBar, mug_query_bar, GTK_TYPE_BOX);
+#else
+G_DEFINE_TYPE (MugQueryBar, mug_query_bar, GTK_TYPE_HBOX);
+#endif /*!HAVE_GTK3*/
+
 
 static void
 mug_query_bar_class_init (MugQueryBarClass * klass)
